#include "NN.h"

#define RAND_6_3 ((((double)(lrand48() % 32767) / 32767) * 2) - 1)

#define MUTATE(x) (((lrand48() % 100) < PROB_MUTATION) ? ((((lrand48() % (DELTA_MUTATION * 2)) - DELTA_MUTATION) * x) / 100.0) + x : x)

int sum ;
int partiegagne;
int nbcoupjoue;
int coup_joue ;

int main(void) {
  int i, j, k;

  int coup;
  int gagnant;
  int meilleur = 0;
  int coupjoue = 0;
  int bas, haut;

  FILE *reseau, *moyenne, *partie, *nbcoup;

  time_t t1, t2;


  t1 = time(NULL);
  srand((unsigned)(time(NULL)));


  reseau = fopen("reseau.dat", "w+");
  moyenne = fopen("moyenne.dat", "w+");
  partie = fopen("partie.dat", "w+");
  nbcoup = fopen("nbcoup.dat", "w+");
 

  // initialisation des poids d'entree
  for (i = 0; i < MAX_POPULATION; i++) {
    for (j = 0; j < (ENTREE + 1); j++) {
      for (k = 0; k < CACHE; k++) {
	poids_entree[i][j][k] = RAND_6_3 ;
      }
    }
  }

  // initialisation des poids de la couche cache
  for (i = 0; i < MAX_POPULATION; i++) {
    for (j = 0; j < CACHE; j++) {
      for (k = 0; k < SORTIE; k++) {
	poids_cache[i][j][k] = RAND_6_3 ;
      }
    }
  }


  // on fait jouer les reseaux les uns contre les autres
  for (k = 0; k < MAX_GENERATION; k++) {
    initialisation_gain();

    sum = 0;
    partiegagne = 0;
    nbcoupjoue = 0;

    for (i = 0; i < MAX_POPULATION; i++) {
      for (j = 0; j < MAX_POPULATION; j++) {
	gagnant = faire_jouer(i, j);
	if (gagnant != -1 ) {
	  partiegagne++;
	  sum += GAIN_GAGNE + GAIN_PERDU;
	  if (gagnant == i) {
	    gain[i] += GAIN_GAGNE;
	    gain[j] += GAIN_PERDU; 
	  }
	  else {
	    gain[j] += GAIN_GAGNE;
	    gain[i] += GAIN_PERDU; 
	  }
	}
	else {
	  sum += GAIN_EGALITE * 2;
	  gain[i] += GAIN_EGALITE;
	  gain[j] += GAIN_EGALITE;
	}
      }
    }

    if (k !=  (MAX_GENERATION - 1)) {
      nouvelle_population();    
    }

    haut = 0;
    bas = 100000;

    for (i=0; i < MAX_POPULATION; i++) {
      if (gain[i] < bas) {
	bas = gain[i];
      }
      if (gain[i] > haut) {
	haut = gain[i];
      }

    }

    fprintf(moyenne, "%f %f %f\n", (double)sum/(MAX_POPULATION * MAX_POPULATION), (double)bas/MAX_POPULATION, (double)haut/MAX_POPULATION);
    fflush(moyenne);
    fprintf(nbcoup, "%f\n", (double)nbcoupjoue/(MAX_POPULATION * MAX_POPULATION) );
    fflush(nbcoup);
    fprintf(partie, "%f\n", (double)partiegagne/MAX_POPULATION);
    fflush(partie);

  }
  fclose(moyenne);
  fclose(nbcoup);
  fclose(partie);

  // choix du meilleur reseau
  for (i = 1; i < MAX_POPULATION; i++) {
    if (gain[i] > gain[meilleur]) {
      meilleur = i;
    }
  }

  t2 = time(NULL);

  printf("Le reseau Choisit est le numero %d en %f secondes\n", meilleur, difftime(t2, t1));
  initialisation_gain();
  initialisation_terrain();



    for (j = 0; j < ENTREE; j++) {
      for (k = 0; k < CACHE; k++) {
	fprintf(reseau, "%f\n", poids_entree[meilleur][j][k] );
      }
    }

  
    for (j = 0; j < CACHE; j++) {
      for (k = 0; k < SORTIE; k++) {
	 fprintf(reseau, "%f\n", poids_cache[meilleur][j][k] );
      }
    }

    fclose(reseau);



    //c la qu'on joue

    while (1) {
      initialisation_terrain();
      coupjoue = 0;
      printf("NOUVELLE PARTIE\n");
      while(1) {
	
	// la machine
	coup = joue(meilleur, 2);
	printf("Coup jou : %d\n", coup);
	affecte_terrain(coup, 2);
	affiche_terrain();
	if (gagne(2) ) {
	  printf ("Perdu\n");
	  break;
	}
	if (++coupjoue >= ENTREE) {
	  printf ("Egalite\n");
	  break;
	}
		// le joueur
	printf("Entrez le numero de la colonne ou jouer\n");
	scanf("%d", &coup);
	DEBUG("Affecte terrain\n");
	affecte_terrain(coup, 1);
	affiche_terrain();
	if (gagne(1) ) {
	  printf ("Gagne\n");
	  break;
	}
	if (++coupjoue >= ENTREE) {
	  printf ("Egalite\n");
	  break;
	}
	

      }
    }


  return 0;
}



int joue(int qui, int quoi) {

  int i, j, k, l;

  double temp_cache, temp_sortie;
  double meilleur = -1.0;
  int coup = 0;


  for (i = 0; i < SORTIE; i++) {
    temp_sortie = 0;
    for (j = 0; j < CACHE; j++) {
      temp_cache = 0;
      for (k = 0; k < HAUTEUR; k++) {
	for (l = 0; l < LARGEUR; l++) {
	  temp_cache +=  (couleur(terrain[l][k], quoi) * poids_entree[qui][(k * LARGEUR) + l][j]  );  
	}
      }
      temp_cache += (drand48() * IMPORTANCE * poids_entree[qui][ENTREE][j] );

      temp_sortie += sigmoid(temp_cache) * poids_cache[qui][j][i];

    }

    if (((abs(temp_sortie) > meilleur) && (terrain[i][HAUTEUR - 1] == 0)) ) {
      meilleur = temp_sortie;
      coup = i;
    }
    
  }
  return coup;
}

int couleur (int pion, int monpion) {
  if (pion) {
    if (pion == monpion) {
      return 1;
    }
    else {
      return 2;
    }
  }
  else {
    return 0;
  }
}  

double sigmoid (double num) {
  //  return (double)(1 / (1 +  exp(-num)));
  return (double)num;
}

void initialisation_terrain(void) {
  int i, j;

  for (i = 0; i < LARGEUR; i++) {
    for (j = 0; j < HAUTEUR; j++) {
      terrain[i][j] = 0;
    }
  } 
}

void affecte_terrain(int coup, int pion) {
  int i;

  for (i = 0; i < HAUTEUR; i++) {
    if (terrain[coup][i] == 0) {
      terrain[coup][i] = pion;
      return;
    }
  }
}

int faire_jouer( int i, int j) {
  int coup;

  coup_joue = 0;

  initialisation_terrain();

  while(1) {

    coup = joue(i, 1);
    affecte_terrain(coup, 1); 

    if (gagne(1)) {
      nbcoupjoue += ++coup_joue;
      return i;
    }
    if (++coup_joue >= ENTREE) {
      nbcoupjoue += coup_joue;
      return -1;
    }

    coup = joue(j, 2);
    affecte_terrain(coup, 2);

   if (gagne(2)) {
      nbcoupjoue += ++coup_joue;
      return j;
    }
    if (++coup_joue >= ENTREE) {
      nbcoupjoue += coup_joue;
      return -1;
    }


  }
}

void initialisation_gain(void) {
  int i;

  for (i = 0; i < MAX_POPULATION; i++) {
    gain[i] = 0;
  }
}

int gagne(int pion) {
  int i, j;


  for (i = 0; i < LARGEUR; i++) {
    for (j = 0; j < HAUTEUR; j++) {
      if (terrain[i][j] == pion) {
	if ((HAUTEUR - j) > 3) {
	  if ((terrain [i][j+1] == pion) && (terrain [i][j+2] == pion) && (terrain [i][j+3] == pion) ) {
	    return 1;
	  }

	  if ((LARGEUR - i) > 3) {
	    if ((terrain [i+1][j+1] == pion) && (terrain [i+2][j+2] == pion) && (terrain [i+3][j+3] == pion) ) {
	      return 1;
	    }
	  }
	}



	if ((LARGEUR - i) > 3) {
	  if ((terrain [i+1][j] == pion) && (terrain [i+2][j] == pion) && (terrain [i+3][j] == pion) ) {
	    return 1;
	  }
	
	  if (i > 2) {
	    if ((terrain [i+1][j-1] == pion) && (terrain [i+2][j-2] == pion) && (terrain [i+3][j-3] == pion) ) {
	      return 1;
	    }
	  }
	}
      }
    }
  }
  return 0;
}

void nouvelle_population(void) {
  double poids_entree_temp[MAX_POPULATION][ENTREE][CACHE];
  double poids_cache_temp[MAX_POPULATION][CACHE][SORTIE];
  
  int i, j, k;
  int max_fitness = 0;
  int select;
  int temp;
  int parent1 = 0, parent2 = 0, parent = 0;
 
  for (i = 0; i < MAX_POPULATION; i++) {
    max_fitness += gain[i];
  }

  max_fitness = (max_fitness == 0 ? 1 : max_fitness);
  
  
  
  for (i = 0; i < MAX_POPULATION; i++) { 
    if ( lrand48()%100 < TAUX_CROSSOVER) {
      
      temp = 0;
      select = lrand48() % max_fitness;
      for (j = 0; j < MAX_POPULATION; j++) {
	if ((temp += gain[j] ) >= select) {
	  parent1 = j;
	  break;
	}
      }
      temp = 0;
      select = lrand48() % max_fitness;
      for (j = 0; j < MAX_POPULATION; j++) {
	if ((temp += gain[j] ) >= select) {
	  parent2 = j;
	  break;
	}
      }
      
      for (j = 0; j < ENTREE; j++) {
	for (k = 0; k < CACHE; k++) {
	  poids_entree_temp[i][j][k] = (((lrand48() % 2) == 0) ? poids_entree[parent1][j][k] : poids_entree[parent2][j][k]) ;
	} 
      } 
      
      for (j = 0; j < CACHE; j++) {
	for (k = 0; k < SORTIE; k++) {
	  poids_cache_temp[i][j][k] = (((lrand48() % 2) == 0) ? poids_cache[parent1][j][k] : poids_entree[parent2][j][k]);
	} 
      }
    }
    
    else {
      temp = 0;
      select = lrand48() % max_fitness;
      for (j = 0; j < MAX_POPULATION; j++) {
	if ((temp += gain[j] ) >= select) {
	  parent = j;
	  break;
	}
      }
      

      for (j = 0; j < ENTREE; j++) {
	for (k = 0; k < CACHE; k++) {
	  poids_entree_temp[i][j][k] = poids_entree[parent][j][k];
	}
      }
      
      for (j = 0; j < CACHE; j++) {
	for (k = 0; k < SORTIE; k++) {
	  poids_cache_temp[i][j][k] = poids_cache[parent][j][k];
	}
      }
      
      if ((lrand48() % 100) < TAUX_MUTATION) {
	for (j = 0; j < ENTREE; j++) {
	  for (k = 0; k < CACHE; k++) {
	    poids_entree_temp[i][j][k] = MUTATE(poids_entree_temp[i][j][k]);
	  }
	}
	
	for (j = 0; j < CACHE; j++) {
	  for (k = 0; k < SORTIE; k++) {
	    poids_cache_temp[i][j][k] = MUTATE(poids_cache_temp[i][j][ k]);
	  }
	}
      }
    }  
  }

  //recopiage dans la liste des genome

  for (i = 0; i < MAX_POPULATION; i++) {
    for (j = 0; j < ENTREE; j++) {
      for (k = 0; k < CACHE; k++) {
	poids_entree[i][j][k] = poids_entree_temp[i][j][k];
      } 
    } 
    
    for (j = 0; j < CACHE; j++) {
      for (k = 0; k < SORTIE; k++) {
	poids_cache[i][j][k] = poids_cache_temp[i][j][k];
      }
    }
    
  }
}


void affiche_terrain(void) {
  int i, j;

  for (j = (HAUTEUR - 1); j >= 0; j--) {
    printf("| ");
    for (i = 0; i < LARGEUR; i++) {
      printf("%d", terrain[i][j]);
    } 
    printf(" |\n");
  } 

  for (i = 0; i < (LARGEUR + 4) ; i++) {
    printf("_");
  } 
  printf("\n");

}

























